
import os

import matplotlib.pyplot as plt
import numpy as np
import scipy.integrate
import scipy.ndimage as ndi

import doas
from smartcarb import io
from smartcarb import misc
from smartcarb import plume


import scipy.optimize as sopt




def distance(q,p):
    return np.sqrt((q[0] - p[0])**2 + (q[1] - p[1])**2)


def find_min_distance_to_curve(p, curve):
    """
    p :: point x0,y0
    f :: function
    c :: coefficients of 2D curve
    """
    constr = [{
        'type': 'eq',
        'fun':  lambda q: curve(x=q[0], y=q[1])[0] - q[0],
    },{
        'type': 'eq',
        'fun':  lambda q: curve(x=q[0], y=q[1])[1] - q[1],
    }]

    q0 = curve(p[0], p[1])

    res = sopt.minimize(distance, x0=q0, args=(p,),
            method='SLSQP', constraints=constr[0],
            options={'maxiter': 100},
    )

    if not res.success:
        print(res)

    return res.x, res.fun, res



def gauss(p, x,y,u, x0=0.0, y0=0.0):
    """\
        Computes plume in kg/m2.

        x: distance from plume (in km)
        y: distance from center line (in km)

        u: wind speed (in meters)
    """
    # conversion factor (kg/m2 -> ppm)   
    air_column = 1.65e29 # molecules per m2
    m = 44.01e-3         # CO2 molar mass: kg/mol
    NA = 6.02214129e23   # 

    c = 1e6 / m / (air_column / NA)

    # km -> m
    x = x * 1e3 # km -> m
    y = y * 1e3 # km -> m

    # state vector
    Q = p[0] / 0.031536  # emissions in Mt/y (convert to kg/s)
    K = p[1]             # eddy diffusion coefficient (in m^2 / s)
    BG = p[2]            # background field (in ppm)

    # dispersion along the plume
    sigma = np.sqrt(2.0 * K * (x-x0) / u)

    # compute Gaussian plume
    A = Q / (np.sqrt(2.0 * np.pi) * sigma * u)
    B = np.exp(-0.5 * (y-y0)**2 / sigma**2)

    p = A * B
    p[x <= 0.0] = 0.0

    # kg/m2 -> ppm
    p = c*p + BG

    return p



def find_all_roots(hh, dd, d, h_max=5000.0):
    """\
    Find all heights (hh), below hh_max where wind direction (dd) equals
    given wind direction (d).
    """

    founds = []

    for i in range(hh.size-1):
        if dd[i] < dd[i+1]:
            dmin, dmax = dd[i], dd[i+1]
            hmin, hmax = hh[i], hh[i+1]
        else:
            dmin, dmax = dd[i+1], dd[i]
            hmin, hmax = hh[i+1], hh[i]

        if hmin > h_max:
            continue

        if dmin <= d < dmax:
            d0 = np.interp(d, [dmin, dmax], [hmin, hmax])
            founds.append(d0)

    return founds




def arc_length_of_poly(f, c, a, b):

    x = np.linspace(a,b)

    s = np.sqrt( 1 + np.polyval(np.polyder(c), x)**2 )

    return scipy.integrate.simps(s,x)


def arc_length_of_2d_curve(curve, a, b):

    t = np.linspace(a,b, 5001)
    xt, yt = curve(t=t, m=1)

    v = np.sqrt( xt**2 + yt**2)

    l = scipy.integrate.simps(v,t)
    return l




def compute_arc_length_and_distance(points, curve):
    """\
    Given n points and polynomial function on [x_o, x_max], compute
    arc length along curve and distance from curve.
    """
    dxs = []
    dys = []

    t_o = curve.get_parameter(curve.x_o, curve.y_o)

    for px,py in points:
        (qx,qy), dy, res = find_min_distance_to_curve([px,py], curve)

        t = curve.get_parameter(qx, qy)
        dx = arc_length_of_2d_curve(curve, t_o, t)

        dxs.append(dx)
        dys.append(dy)

    return np.array(dxs), np.array(dys)





def flux_inversion(p0, dx, dy, field, wind, std=0.7):

    def model(x,b):
        return gauss(x,*b)

    p0 = np.asarray(p0)
    Se = np.full(field.shape, std**2)

    b = (dx,dy,wind)
    x, res = doas.solver.gauss_newton(p0, field, model, Se=Se, b=b)

    return x, res



class Poly2D:
    def __init__(self, x, y, w, degree=2, x_o=0.0, y_o=0.0,
            x0=0.0, y0=0.0):
        """\
        A 2D curve fitted on the point cloud given by x and y.

        Parameters
        ----------
        x,y,w: x,y coords and weights used for fitting the data

        degree: degrees if the two polynomials x(t) and y(t)

        x_o, y_o: location of source (will be added to x,y and given
                  high weight such that curves goes through source)

        x0, y0: origin of coordinate system
        """
        self.x = x
        self.y = y
        self.w = w
        self.degree = degree

        self.x_o = x_o
        self.y_o = y_o

        self.x0 = x0
        self.y0 = y0

        self.c = np.zeros(2 * (degree + 1))

        self._fit()


    def _fit(self):

        def objective(c, w, x, y):
            xt,yt = self._compute_curve(c,x,y)
            return np.concatenate([
                w * (xt - x), w * (yt - y)
            ])

        # add origin
        x = np.append(self.x, self.x_o)
        y = np.append(self.y, self.y_o)
        w = np.append(self.w, 1000.)

        # angle around origin (not used)
        #phi = np.arctan2(y - y0, x - x0)

        # curve fit
        self.c, res = sopt.leastsq(objective, x0=self.c, args=(w,x,y))

        if res not in [1,2,3,4]:
            print('least square failed with error code: ', res)


    def get_parameter(self, x, y):
        return np.sqrt((x - self.x0)**2 + (y - self.y0)**2)


    def _compute_curve(self, c, x=None, y=None, t=None, m=0):
        if t is None:
            t = self.get_parameter(x,y)

        k = c.size // 2
        cx = c[:k]
        cy = c[k:]

        if m != 0:
            cx = np.polyder(cx, m)
            cy = np.polyder(cy, m)

        xt = np.polyval(cx,t)
        yt = np.polyval(cy,t)

        return xt, yt

    def __call__(self, x=None, y=None, t=None, m=0):
        return self._compute_curve(self.c, x, y, t, m)



def transform_to_kilometers(lon, lat, lon_o, lat_o, epsg=3006):
    """\
    Transform WGS84 to meters (using EPSG system).
    """
    x_o, y_o = misc.transform_points(lon_o, lat_o, epsg)
    x, y = misc.transform_points(lon, lat, epsg)
    return (x - x_o) / 1e3, (y - y_o) / 1e3


def setup_flux_inversion(lon0, lat0, field, clouds, is_plume, t_values, noise,
        lon_model, lat_model, height_model, uwind, vwind,
        origin=None, lon_o=None, lat_o=None,
        degree=2, epsg=3006,
        **kwargs):
    """
    LEVEL-2 data
    ------------
    lon0, lat0 :: center longitude/latitude of ground pixels
    field      :: satellite observations (e.g. XCO2) - without noise
    clouds     :: cloud fractions at ground pixels
    is_plume   :: boolean array where plume has been detected
    t_values   :: t-values from plume detection
    noise      :: noise on field

    MODEL DATA (e.g COSMO)
    ----------------------
    lon_model    :: longitude in model (lon, lat)
    lat_model    :: latitude in model (lon, lat)
    height_model :: height in model (lon, lat, layers+1)
    uwind        :: u-wind in model (lon, lat, layers)
    vwind        :: v-wind in model (lon, lat, layers)

    FLUX INVERSION PARAMETERS
    -------------------------
    origin       :: name of plume (used to get lon_o and lat_o)
    lon_o, lat_o :: longitude/latitude of point source
    degree       :: degree of polynomial fitted on plume

    """
    # TRANSFORM LON/LAT TO CARTESIAN COORDS
    if origin is not None:
        lon_o, lat_o = plume.get_source_origin(origin)

    i_o, j_o = misc.nearest_index(lon_o, lat_o, lon0, lat0)
    x_o, y_o = 0.0, 0.0

    x_sat, y_sat = transform_to_kilometers(lon0, lat0, lon_o, lat_o, epsg)
    x_model, y_model = transform_to_kilometers(lon_model, lat_model, lon_o, lat_o, epsg)


    # DETERMINE CENTER LINE OF PLUME 
    # get location of plume and XCO2 values (plus wind etc.)
    x_p = x_sat[is_plume]
    y_p = y_sat[is_plume]
    t_p = t_values[is_plume]

    # origin of center line
    x0, y0 = -x_p.mean(), -y_p.mean()
    print('center line origin:', x0, y0)

    # fit 2d curve
    curve = Poly2D(x_p, y_p, t_p, degree, x_o, y_o, x0, y0)
    t = curve.get_parameter(x_p, y_p)
    t = np.linspace(t.min(), t.max())
    x_curve, y_curve = curve(t=t)

    # DISTANCE FROM CENTER LINE
    # make plume bigger to include background
    struct = ndi.generate_binary_structure(2, 2)
    full_area = ndi.binary_dilation(is_plume, struct, iterations=3) # TODO iterations?

    # TODO
    full_area &= (clouds <= 0.01)

    # compute for all ground pixels distance to center line
    x_a = x_sat[full_area]
    y_a = y_sat[full_area]
    field_a = field[full_area] + noise[full_area]


    points = np.transpose([x_a, y_a])
    dx, dy = compute_arc_length_and_distance(points, curve)


    # remove points downstream of lon_p.max()
    x_a = x_sat[full_area][np.isfinite(dx)]
    y_a = y_sat[full_area][np.isfinite(dx)]
    field_a = field[full_area][np.isfinite(dx)] + noise[full_area][np.isfinite(dx)]

    is_plume2 = is_plume[full_area][np.isfinite(dx)]

    dy = dy[np.isfinite(dx)]
    dx = dx[np.isfinite(dx)]

    ###################################################
    # WIND SPEED ALONG CENTER LINE (TODO)             #
    # (compute mean wind speed below 2000m at source) #
    ###################################################

    i_o, j_o = misc.nearest_index(x_o, y_o, x_model, y_model)

    # wind profile at point source (TODO: spatial interpolation)
    uu = uwind[::-1, i_o, j_o]
    vv = vwind[::-1, i_o, j_o]
    hh = height_model[::-1, i_o, j_o]
    hh = 0.5 * (hh[1:] + hh[:-1])
    dd = (270.0 - np.rad2deg(np.arctan2(vv,uu))) % 360

    # wind direction (satellite)
    xt,yt = curve(x_o, y_o, m=1)
    d_sat = (270 - np.rad2deg(np.arctan2(yt,xt))) % 360

    # mean wind
    wind = np.sqrt(np.mean(uu[hh <= 2000])**2 + np.mean(vv[hh <= 2000])**2)
    print('mean wind:', wind)

    # wind from satellite wind direction
    wind = estimate_wind_speed(hh, uu, vv, d_sat)

    # TODO: compare estimated height with emission height in emission files
    print('plume length:\t%d km' % dx.max())
    print('plume age:\t%.1f hours' % (1e3 * dx.max() / wind / 3600))


    return x_curve, y_curve, x_a, y_a, field_a, dx, dy, wind, \
            curve, is_plume2, hh,uu,vv,dd,d_sat



def estimate_wind_speed(hh,uu,vv,d_sat, hmin=0, hmax=2000.0):

    dd = (270.0 - np.rad2deg(np.arctan2(vv,uu))) % 360

    # brute-force match wind direction between plume and model
    k_o = np.argmin(
            misc.anglediff(dd[(hmin <= hh) & (hh < hmax)], d_sat)**2
    )
    wind = np.sqrt(uu[k_o]**2 + vv[k_o]**2)

    print('first estimate:', hh[k_o], wind)

    # fine-tuning
    hh2 = np.linspace(hh[k_o-1], hh[k_o+1], 51)
    uu2 = np.interp(hh2, hh, uu)
    vv2 = np.interp(hh2, hh, vv)
    dd2 = (270.0 - np.rad2deg(np.arctan2(vv2,uu2))) % 360

    k2 = np.argmin(misc.anglediff(dd2, d_sat)**2)

    wind = np.sqrt(uu2[k2]**2 + vv2[k2]**2)

    print('fine estimate:', hh2[k2], wind)

    return wind



def visualize(x_c, y_c, x_a, y_a, f_a, x, dx, dy, wind, is_plume, filename):

    v = round(np.mean(f_a))
    s = max(1, round(np.std(f_a)))
    vmin, vmax = v-s, v+s

    fig = plt.figure()

    ax1 = fig.add_subplot(121, aspect='equal')
    ax2 = fig.add_subplot(122, aspect='equal')


    ax1.set_ylabel('y (km)')
    ax1.set_title('(a) Observations', fontsize=10)
    ax2.set_title('(b) Fit', fontsize=10)


    c = ax1.scatter(x_a, y_a, c=f_a, vmin=vmin, vmax=vmax)
    c = ax2.scatter(x_a, y_a, c=gauss(x,dx,dy, u=wind), vmin=vmin, vmax=vmax)
    ax1.scatter(x_a[is_plume], y_a[is_plume], edgecolors='#222222', c='none')

    for ax in [ax1, ax2]:
        ax.set_xlabel('x (km)')
        ax.plot(x_c, y_c, 'k-')

        # draw arrow
        d1,d2 = x_c[-1] - x_c[-2], y_c[-1] - y_c[-2]
        d1 = d1 / np.sqrt(d1**2 + d2**2)
        d2 = d2 / np.sqrt(d1**2 + d2**2)
        ax.arrow(x_c[-1], y_c[-1], d1, d2, width=0, lw=1, head_width=2,
                head_length=2, fc='k', ec='k')

        ax.grid(True)

    label = io.filename2label(filename)
    text = ', '.join([
        'Q = %.1f Mt/y' % x[0],
        'K = %d m$^2$/s' % x[1],
        'x$_{BG}$ = %d ppm' % x[2],
        'u = %.1f m/s' % wind
    ])
    fig.suptitle('Gaussian plume fit for Janschwalde: %s\n(%s)' % (label, text), fontsize=10, y=0.99)



    fig.subplots_adjust(right=0.8)
    cax = fig.add_axes([0.82, 0.15, 0.04, 0.7])
    cb = fig.colorbar(c, cax=cax, extend='both')
    cb.set_label('XCO$_2$ (ppm)')


    # text: wind-speed


def estimate_plume_height(z, d, d0, h0=100):
    """\
    Estimate height of plume using wind direction of plume d0 and
    profile of wind direction d(z). The algorithm will minimize the difference
    between them.
    """
    def f(h, z, d, d0):
        d = np.interp(h,z,d)
        return misc.anglediff(d0,d)**2

    #TODO: does not work very well because of several local minima
    res = sopt.minimize(f, x0=h0, args=(z,d,d0))

    return res


def estimate_wind_speed_along_curve(xc, yc, x_model, y_model, height_model, uwind, vwind,
        f_curve, c_curve, x0, y0):

    heights = []
    speed = []
    direction = []
    direction_sat = []

    uu = []
    vv = []

    for x,y in zip(xc[::4],yc[::4]):
        i_o, j_o = misc.nearest_index(x, y, x_model, y_model)

        # estimate cardinal wind direction (and plume height) 
        # along center line (TODO)
        uwind2 = uwind[::-1, i_o, j_o][:30]
        vwind2 = vwind[::-1, i_o, j_o][:30]
        hh = height_model[::-1, i_o, j_o]
        hh = 0.5 * (hh[1:] + hh[:-1])[:30]

        heights.append(hh)
        uu.append(uwind2)
        vv.append(vwind2)

        u_mean = np.mean(uwind2[(200 <= hh) & (hh < 1000.0)])
        v_mean = np.mean(vwind2[(200 <= hh) & (hh < 1000.0)])

        speed.append( np.sqrt(uwind2**2 + vwind2**2))

        # wind direction (model)
        dd = (270.0 - np.rad2deg(np.arctan2(uwind2,vwind2))) % 360
        direction.append(dd)

        # wind direction (satellite)
        xt,yt = f_curve(c_curve, x=x, y=y, m=1, x0=x0, y0=y0)
        d_sat = (270 - np.rad2deg(np.arctan2(yt,xt))) % 360

        direction_sat.append(d_sat)

    return xc, yc, np.array(heights), np.array(speed), np.array(direction), \
            np.array(direction_sat), np.array(uu), np.array(vv)










def main(origin, p_thr=0.95, footprint=1):

    if origin in ['Berlin', 'Lippendorf', 'Patnow']:
        fn = 'Level-2_%s.dat' % origin
    else:
        fn = 'Level-2_Others.dat'

    path = '/users/gkuhl/projects/smartcarb/plume_detection/'
    filenames = np.loadtxt(os.path.join(path, fn), dtype=str)


    for filename in filenames:

        path = '/scratch/snx3000/gkuhl/smartcarb/level2/Sentinel_7_CO2/'
        filename = os.path.join(path, filename)


        is_plume, info = plume.plume_detection_algorithm(filename, tracer='CO2',
                seed=None, use_epf=True, error_std=0.7,
                bg_type='model', bg_area=25, origin=origin,
                p_thr=p_thr, footprint=footprint, cloud_thr=0.01,
                show=False
        )

        print(os.path.basename(filename), info['mean_bg'],
                np.sqrt(info['var_bg']), info['n_bg'])


        #lon,lat,U,V = io.read_3D_for_level2(filename, ['lon', 'lat', 'U', 'V'], '3D')
        #HHL = io.read_3D_for_level2(fnames[i], ['HHL'], '3Dc')[0]
























