#!/usr/bin/env python
# coding: utf-8

from datetime import datetime, timedelta

import numpy as np
import scipy.ndimage as ndi
import scipy.optimize as opt

import doas

from smartcarb import io
from smartcarb import misc
from smartcarb import plume


# kg/s -> Mt/y (Tg/y)
kgs_to_Mty = 1e3 * (365 * 24 * 60 * 60) / 1e12


def compute_emissions(nc):

    cats = [
       'CO2_point_snap1_major', # this was counted twice in Jan and Jul
       'CO2_point_snap1_minor',
       'CO2_point_snap2_minor',
       'CO2_point_snap3_minor',
       'CO2_point_snap4_minor',
       'CO2_point_snap5_minor',
       'CO2_point_snap34_minor',
       'CO2_point_snap8_minor',
       'CO2_point_snap9_minor',
       'CO2_area_construction',
       'CO2_line_aircraft',
       'CO2_point_aircraft',
       'CO2_area_aircraft',
       'CO2_area_forestry',
       'CO2_area_offroad',
       'CO2_line_shipping',
       'CO2_point_shipping',
       'CO2_line_train',
       'CO2_area_heating',
       'CO2_line_majorroads',
       'CO2_area_minorroads'
    ]

    total = 0.0
    for cat in cats:
        total += np.sum(nc.variables[cat][:] * 1235941.1)

    return total



def get_constant_emissions(time, use_constant=True):
    if not use_constant and time.year == 2015 and time.month in [1,7]:
        return 754.86
    else:
        return 534.30



def simple_leastsq(time, filename=None, use_level2=True,
        tracer='CO2', correct_berlin=True,
        instrument='Sentinel_7_CO2', orbit=0, lon_eq=0,
        data_path='.'
    ):
    """\
    filename is either a file with 2d COSMO field or Level-2 data

    y[i,j] = x * H[i,j] + b

    - y: XCO2 measurements in plume
    - H: XCO2 from Berlin tracer (e.g. XCO2_BC) normalized by emissions
    - b: background (e.g. mean of "background" tracers in COSMO)

    The plume is defined where 'plume_tracer' > 'plume_thr'.

    If `use_const_emissions` use XCO2_BV in observation vector y.
    """

    if time >= datetime(2015,7,1):
        starttime = datetime(2015,7,1)
    else:
        starttime = datetime(2015,1,1)

    if filename is None:
        input_files = 'level2' if use_level2 else '2D'
        filename = io.create_filename(time, starttime, input_files,
                instrument, orbit, lon_eq, data_path
        )

    # longitude/latitude
    lon, lat = io.read_fields(filename, ['longitude', 'latitude'])

    if (np.min((lat - cfg.BERLIN[1])**2 + (lon - cfg.BERLIN[0])**2)) > 0.5:
        return np.nan, np.nan, {'number of pixels': np.nan}

    # tracer fields (synthetic observations)
    co2 = io.read_total_tracer(filename, 'CO2', time, correct_berlin)

    if tracer == 'NO2':
        no2 = io.read_total_tracer(filename, 'NO2', time, correct_berlin)

    if not use_level2:
        # spatial binning (for 2D fields)
        co2 = misc.spatial_binning(co2)

        if tracer == 'NO2':
            no2 = misc.spatial_binning(no2)


    # TODO: read noise


    # plume detection (for CO2 plume)
    is_plume, _ = plume.plume_detection_algorithm(...) # TODO



    info['number of pixels'] = np.sum(is_plume)




    if np.sum(is_plume) == 0:
        return np.nan, np.nan, info

    if background_mode == 'model':
        bg_std = 0.0
    elif background_mode == 'mean':
        bg_std = b.std()
        b = b.mean()
    else:
        raise ValueError('unknown mode for background estimation')

    # observation vector
    if use_const_emissions:
        y = (bc + jv + anth + ra - gpp + bg)[is_plume]
    else:
        y = (bv + jv + anth + ra - gpp + bg)[is_plume]

    # footprint
    H = bc[is_plume] / get_constant_emissions(time)

    # add noise
    y += sat_noise * np.random.randn(y.size)

    # Se, xa and Sa
    if sat_noise < 1e-16:
        Se = np.nan
    else:
        Se = np.full(y.shape, sat_noise**2 + bg_std**2)


    return y, Se, H, b




def flux_inversion(H, y, Se, b):
    """\
        H  :: footprint
        y  :: observations
        Se :: uncertainty
        b  :: background field

    """
    # forward model
    f = lambda x,_: H * x + b

    x0 = np.array([534.0])
    x, res = doas.solver.gauss_newton(x0, y, f, Se=Se)

    return x[0], np.sqrt(float(res['S'])), info






