#!/usr/bin/env python
# coding: utf-8

from __future__ import print_function, division
from six import string_types

from datetime import datetime, timedelta
import glob
import os

import numpy as np
import netCDF4
import xarray

from smartcarb import config as cfg
from smartcarb import misc


def process_filename(filename):
    #Sentinel_7_CO2_2015051511_o2009_l270.nc
    filename = os.path.basename(filename)
    filename = os.path.splitext(filename)[0]
    sat, n, t, date, orbit, lon = filename.split('_')[:6]

    sat = '_'.join([sat, n, t])
    orbit = int(orbit[1:])
    date = datetime.strptime(date, '%Y%m%d%H')
    lon = int(lon[1:])

    return sat, orbit, date, lon


def filename2label(filename):
    sat, orbit, date, lon = process_filename(filename)

    letter = 'ABCDEF'[lon//90]
    sat = 'CO$_2$ satellite %s' % letter

    label = '%s (orbit #%d at %s)' % (sat, orbit,
            date.strftime('%Y-%m-%d %H:%M UTC'))

    return label


def filename2time(filename):
    #Sentinel_7_CO2_2015051511_o2009_l270.nc
    return datetime.strptime(
        os.path.basename(filename).split('_')[3], '%Y%m%d%H'
    )




def str2char(s):
    return np.array(list(s), 'c')


def init_netcdf(nc, rlon, rlat, lon=None, lat=None, pollon=cfg.pollon, pollat=cfg.pollat,
        attrs=None):
    """\
    Init netCDF4 file with cosmo rotated pole grid
    """
    if attrs is not None:
        nc.setncatts(attrs)

    # dimensions
    nc.createDimension('time')
    nc.createDimension('rlat', rlat.size)
    nc.createDimension('rlon', rlon.size)

    # time variable
    nc_time = nc.createVariable('time', 'f8', ('time',))
    nc_time.setncattr('units', str2char('seconds since 2015-01-01 00:00'))
    nc_time.setncattr('calendar', str2char('proleptic_gregorian'))

    # rlon
    nc_rlon = nc.createVariable('rlon', 'f8', ('rlon',))
    nc_rlon[:] = rlon
    nc_rlon.setncattr('axis', str2char('X'))
    nc_rlon.setncattr('units', str2char('degrees'))
    nc_rlon.setncattr('standard_name', str2char('longitude'))

    # rlat
    nc_rlat = nc.createVariable('rlat', 'f8', ('rlat',))
    nc_rlat[:] = rlat
    nc_rlat.setncattr('axis', str2char('Y'))
    nc_rlat.setncattr('units', str2char('degrees'))
    nc_rlat.setncattr('standard_name', str2char('latitude'))

    # rotated pole
    nc_pole = nc.createVariable('rotated_pole', 'c', ('time','rlat','rlon'))
    nc_pole.setncattr('grid_mapping_name', str2char('rotated_latitude_longitude'))
    nc_pole.setncattr('grid_north_pole_latitude', pollat)
    nc_pole.setncattr('grid_north_pole_longitude', pollon)
    nc_pole.setncattr('north_pole_grid_longitude', 0.0)

    # longitude
    if lon is not None:
        nc_lon = nc.createVariable('lon', 'f8', ('rlat', 'rlon'))
        nc_lon[:] = lon
        nc_lon.setncattr('axis', str2char('X'))
        nc_lon.setncattr('units', str2char('degrees'))
        nc_lon.setncattr('standard_name', str2char('longitude'))

    # latitude
    if lat is not None:
        nc_lat = nc.createVariable('lat', 'f8', ('rlat', 'rlon'))
        nc_lat[:] = lat
        nc_lat.setncattr('axis', str2char('Y'))
        nc_lat.setncattr('units', str2char('degrees'))
        nc_lat.setncattr('standard_name', str2char('latitude'))



def append_variable(nc, name, values, attrs=None):
    """\
    Append variable (name, values) to netCDF file.
    """
    if attrs is None:
        attrs = {}

    var = nc.createVariable(name, values.dtype, ('time', 'rlat', 'rlon'), zlib=True)
    var[0] = values

    # add or update attributes
    for key, value in attrs.items():
        if isinstance(value, string_types):
            var.setncattr(key, str2char(value))
        else:
            var.setncattr(key, value)


def get_slice(origin, lon=None, lat=None, area=100):
    """\
    e.g. Berlin, area=40
    e.g. Janschwalde, area=30
    """

    if lon is None or lat is None:
        lon, lat = misc.rotpole_to_wgs84(cfg.rlon2, cfg.rlat2, cfg.pollon, cfg.pollat)

    if origin in cfg.CITIES:
        lon_o, lat_o = cfg.CITIES[origin]
    else:
        lon_o, lat_o = cfg.POINT_SOURCES[origin]

    i,j = misc.nearest_index(lon_o, lat_o, lon, lat)

    return slice(i-area//2,i+area//2+1), slice(j-area//2,j+area//2+1)


def list_level2_filenames(filename, path='.'):

    fnames = []

    with open(filename) as list_file:
        for line in list_file:
            if line.startswith('#'):
                continue

            date, sat = line.strip().split()
            lon = {'1': 0, '2': 90, '3': 180, '4': 270}[sat]
            date = date.replace('-', '')

            pattern = os.path.join(path, '*_%s??_o????_l%03d.nc' % (date, lon))

            fnames += glob.glob(pattern)

    return fnames





def create_filename(time, starttime=None, input_files='3D', instrument='', orbit=0,
        lon_eq=0, path=cfg.DATA_PATH, use_restart_structure=False,
    ):
    """\
    Create filename for netCDF file.
    """
    if starttime is None:
        starttime = time2starttime(time)

    if input_files in ['3Dc', '3D']:
        hstart = (time - starttime).days // 3 * 72
        hstop = hstart + 72

        if use_restart_structure:
            path = os.path.join(
                    path,
                    '%s_%d_%d' % (starttime.strftime('%Y%m%d%H'), hstart, hstop),
                    'cosmo_output'
            )

        if input_files == '3D':
            filename = os.path.join(
                    path, time.strftime('lffd%Y%m%d%H.nc')
            )
        else:
            filename = os.path.join(
                    path, starttime.strftime('lffd%Y%m%d%Hc.nc')
            )

    elif input_files == '2D':
        path = os.path.join(path, starttime.strftime('cosmo_2D_%Y-%m'))
        filename = os.path.join(
                path, time.strftime('cosmo_2d_%Y%m%d%H.nc')
        )

    elif input_files == 'level2':
        # Level-2 data for instrument, orbit and lon_eq

        if not isinstance(orbit, string_types):
            orbit = '%04d' % orbit
        if not isinstance(lon_eq, string_types):
            lon_eq = '%03d' % lon_eq

        pattern = '{0:s}_%Y%m%d??_o{1:s}_l{2:s}.nc'.format(instrument, orbit, lon_eq)
        pattern = os.path.join(path, 'level2', instrument, time.strftime(pattern))

        filename = glob.glob(pattern)

        if not filename:
            print(pattern)
        if len(filename) > 1:
            print('Use first file of %d files.' % len(filename))
        filename = filename[0]

    return filename



def read_3D_for_level2(level2_filename, fields, type_='3D', path=None,
        correct_berlin=False, slice_=Ellipsis):
    """
    Read 3D data for Level-2 data.
    """
    time = process_filename(level2_filename)[2]
    starttime = time2starttime(time)

    if path is None: # TODO: move to config
        path = starttime.strftime('/project/d73/output/cosmo_%Y-%m/')

    filename = create_filename(time, starttime, type_, path=path)

    return read_fields(filename, fields, time, correct_berlin, slice_)


def read_ground_pixels(filename, slice_=Ellipsis):
    """Read center and corners of ground pixels."""
    return read_fields(filename, [
            'longitude', 'latitude', 'longitude_corners', 'latitude_corners'
        ], slice_=slice_)



def read_total_tracer(filename, tracer, time=None, correct_berlin=False,
        clip_data=True, slice_=Ellipsis):
    """\
    Read total tracer field (CO2, NO2 or CO).
    """
    if tracer in ['CO2', 'XCO2']:
        fields = ['XCO2_BV', 'XCO2_A', 'XCO2_JV', 'XCO2_RA', 'XCO2_GPP', 'XCO2_BG']
        bv,a,jv,ra,gpp,bg = read_fields(filename, fields, time, correct_berlin, slice_)

        field = bv+a+jv+ra-gpp+bg

        if not hasattr(field, 'mask'):
            field = np.ma.array(field, mask=np.zeros(field.shape, bool))

        if clip_data:
            field.mask[(field < 0) | (field > 1e5)] = True

        return field

    if tracer == 'NO2':
        fields = ['NO2_BV', 'NO2_A', 'NO2_JV', 'NO2_BG']
        bv,a,jv,bg = read_fields(filename, fields, time, correct_berlin, slice_)
        return bv+a+jv+bg

    if tracer == 'CO':
        fields = ['CO_BV', 'CO_A', 'CO_JV', 'CO_BG']
        bv,a,jv,bg = read_fields(filename, fields, time, correct_berlin, slice_)
        return bv+a+jv+bg


def read_tracer_noise(filename, tracer, use_epf=True, level=0.7, slice_=Ellipsis):
    """
    tracer in {'CO2', 'NO2', 'CO', 'uXCO2', 'uNO2_low', 'uNO2_high',
               'uCO_low', 'uCO_high'}
    level in {0.5, 0.7, 1.0, 'low', 'high'}
    """

    if tracer == 'CO2':
        tracer = 'uXCO2'
    elif tracer == 'NO2':
        tracer = 'uNO2_%s' % level
    else:
        tracer = 'uCO_%s' % level

    # read uncertainty (TODO/FIXME: correct uNO2 for Jan/Jul)
    u = read_fields(filename, [tracer], slice_=slice_)[0]

    noise = np.random.randn(*u.shape)

    if tracer == 'uXCO2':
        if use_epf:
            noise = level * noise * u / 1.50
        else:
            noise = level * noise

    elif tracer == 'NO2' and use_epf:
        cc = read_fields(filename, ['CLCT'], slice_=slice_)[0]

        # use Wenig et al. (2008) with `uNO2` instead of 1.5e15
        noise = (1.0 + 3.0 * cc) * u * noise

    else:
        noise = u * noise

    return noise





def read_fields(filename, tracers, time=None, correct_berlin=False,
        slice_=Ellipsis):
    """\
    Read fields from netCDF file. If correct_berlin is True,
    correct '*_BV' tracers for January and July
    by scaling '*_B0' with 0.55 which corrects the too high
    emissions in January and July in the SMARTCARB simulations.
    """
    with netCDF4.Dataset(filename) as nc:
        fields = []

        for tracer in tracers:
            if correct_berlin and tracer.endswith('_BV'):
                label = tracer.split('_')[0]

                berlin_tracers = ['%s_B%d' % (label, i) for i in range(3)]
                b0,b1,b2 = read_fields(filename, berlin_tracers, time,
                        correct_berlin, slice_)

                if time.month in [1,7]:
                    field = 0.55 * b0 + b1 + b2
                else:
                    field = b0 + b1 + b2
            else:
                field = nc.variables[tracer]

                if slice_ is Ellipsis or isinstance(slice_, slice):
                    field = field[slice_]
                else:
                    # use slice starting at last dimension (instead of first)
                    ndim = np.ndim(field)

                    if len(slice_) == ndim:
                        field = field[slice_]
                    else:
                        s = [Ellipsis] * (ndim - len(slice_)) + list(slice_)
                        field = field[s]

            field = np.squeeze(field)
            fields.append(field)

    return fields




def read_field(time, starttime, keys, indices=Ellipsis, input_files='3D',
        path='.', instrument='', orbit=0, lon_eq=0
    ):
    if indices is Ellipsis:
        indices = [Ellipsis for _ in keys]


    if input_files in ['3Dc', '3D']:
        hstart = (time - starttime).days // 3 * 72
        hstop = hstart + 72

        folder = '%s_%d_%d' % (starttime.strftime('%Y%m%d%H'), hstart, hstop)

        if input_files == '3D':
            filename = os.path.join(
                    path, folder, 'cosmo_output',
                    time.strftime('lffd%Y%m%d%H.nc')
            )
        else:
            filename = os.path.join(
                    path, folder, 'cosmo_output',
                    starttime.strftime('lffd%Y%m%d%Hc.nc')
            )

    elif input_files == '2D':
        path = os.path.join(path, starttime.strftime('cosmo2D_%Y-%m'))
        filename = os.path.join(
                path, time.strftime('cosmo_2d_%Y%m%d%H.nc')
        )

    # read Level-2 data for instrument, orbig and lon_eq
    elif input_files == 'level2':

        filename = '{0:s}_%Y%m%d??_o{1:04d}_l{2:03d}.nc'.format(instrument, orbit, lon_eq)
        filename = os.path.join(path, 'level2', instrument, time.strftime(filename))

        filename = glob.glob(filename)
        if len(filename) > 1:
            print('Use first file of %d files.' % len(filename))
        filename = filename[0]


    print(filename)
    with netCDF4.Dataset(filename) as nc:
        data = [nc.variables[key][index] for key, index in zip(keys, indices)]

    return data[0] if len(data) == 1 else data


def time2starttime(time):
    if time >= datetime(2015,7,1):
        starttime = datetime(2015,7,1)
    else:
        starttime = datetime(2015,1,1)
    return starttime



def read_signal_and_background(filename, tracer, time=None,
        correct_berlin=False, slice_=Ellipsis):

    signal = read_fields(filename, [tracer], time=time,
            correct_berlin=correct_berlin, slice_=slice_)[0]
    totals = read_total_tracer(filename, tracer.split('_')[0], time=time,
            correct_berlin=correct_berlin, slice_=slice_)
    bg = totals - signal

    return signal, bg



def read_signal_and_background_old(time, tracer, area, input_files,
        instrument='', orbit=0, lon_eq=0, nc_filename=None):
    """\
    """
    import warnings
    raise warnings.DeprecationWarning('do not use this function')
    if area == 'Berlin':
        ref = 'BV'
        oth = 'JV'
    else:
        ref = 'JV'
        oth = 'BV'

    starttime = time2starttime(time)

    if nc_filename is None:
        nc_filename = create_filename(time, starttime, input_files,
            instrument, orbit, lon_eq, path=cfg.ROOT
        )

    if tracer == 'CO2':
        signal, clct, bg, anth, jv, gpp, ra, noise = read_fields(nc_filename,
            ['XCO2_%s' % ref, 'CLCT', 'XCO2_BG', 'XCO2_A',
             'XCO2_%s' % oth, 'XCO2_GPP', 'XCO2_RA', 'uXCO2'],
            time=time, correct_berlin=True
        )
        background = bg + anth + jv - gpp + ra

    elif tracer == 'NO2':
        signal, clct, bg, anth, jv, noise = read_fields(nc_filename,
                ['NO2_%s' % ref, 'CLCT', 'NO2_BG', 'NO2_A',
                 'NO2_%s' % oth, 'uNO2_low'],
            time=time, correct_berlin=True
        )
        background = bg + anth + jv

    elif tracer == 'CO':
        signal, clct, bg, anth, jv, noise = read_fields(nc_filename,
            ['CO_%s' % ref, 'CLCT', 'CO_BG', 'CO_A', 'CO_%s' % oth, 'uCO_low'],
            time=time, correct_berlin=True
        )
        background = bg + anth + jv

    # spatial binning
    if input_files != 'level2':
        signal = misc.spatial_binning(signal)[0]
        clct = misc.spatial_binning(clct)[0]
        background = misc.spatial_binning(background)[0]

    mask = np.isnan(signal)

    return np.ma.array(signal, mask=mask), np.ma.array(clct, mask=mask), \
           np.ma.array(background, mask=mask), np.ma.array(noise, mask=mask)




if __name__ == '__main__':
    pass


