#!/usr/bin/env python
# coding: utf-8

from datetime import datetime, timedelta

import osr, ogr
import numpy as np
import scipy
import scipy.ndimage as ndi
import cartopy.crs as ccrs


def anglediff(alpha, beta, is_degree=True):
    if is_degree:
        alpha = np.deg2rad(alpha)
        beta = np.deg2rad(beta)

    delta = np.arctan2(
        np.sin(alpha-beta),
        np.cos(alpha-beta)
    )
    return np.rad2deg(delta) if is_degree else delta


def nearest_index(lon_o, lat_o, lon, lat):
    index = np.argmin(np.abs(lon - lon_o) + np.abs(lat - lat_o))
    i,j = np.unravel_index(index, lon.shape)

    return i,j


def wind_direction(u,v):
    return (270.0 - np.rad2deg(np.arctan2(v,u))) % 360


def iter_times(start, end, step=timedelta(hours=1), exclude_end=False):
    current = start

    if exclude_end:
        while current < end:
            yield current
            current = current + step
    else:
        while current <= end:
            yield current
            current = current + step



def create_rotpole_reference(pollon, pollat):
    """\
    Return spatial reference for rotated pole coordinates.

    pollon: longitude of north pole
    pollat: latitude of north pole
    """
    ref = osr.SpatialReference()
    ref.ImportFromEPSG(4326)     # WSG-84

    if (pollon is not None) and (pollat is not None):
        o_lon_p = 0.0
        o_lat_p = pollat
        lon_0 = 180 + pollon

        string = (
            'PROJCS["rotated_pole projection",EXTENSION["PROJ4","+proj=ob_tran +o_proj=longlat'
            ' +o_lon_p=%.2f +o_lat_p=%.2f +lon_0=%.2f +a=1 +to_meter=0.0174532925199"]]'
        ) % (o_lon_p, o_lat_p, lon_0)

        ref.ImportFromWkt(string)    # rotpole transformation

    return ref



def transform_points(lon, lat, epsg=3006):
    """
    """
    out = ccrs.epsg(epsg)
    in_ = ccrs.PlateCarree()

    if np.ndim(lon) == 0:
        res = out.transform_point(lon, lat, in_)
        return res[0], res[1]
    else:
        res = out.transform_points(in_, lon, lat)
        return res[...,0], res[...,1]


def epsg2rotpole(epsg, pollon, pollat, inverse=False):
    """\
    Create transformation from EPSG to rotpole described
    by pollon, pollat.
    """
    ref0 = osr.SpatialReference()
    ref0.ImportFromEPSG(epsg)

    ref1 = create_rotpole_reference(pollon, pollat)

    if inverse:
        return osr.CoordinateTransformation(ref1, ref0)
    else:
        return osr.CoordinateTransformation(ref0, ref1)




def rotpole_to_wgs84(rlon, rlat, pollon, pollat):

    c_in = ccrs.RotatedPole(pollon, pollat)
    c_out  = ccrs.PlateCarree()

    if np.ndim(rlon) == 0:
        res = c_out.transform_point(rlon, rlat, c_in)
        return res[0], res[1]
    else:
        res = c_out.transform_points(c_in, rlon, rlat)
        return res[...,0], res[...,1]




def wgs84_to_rotpole(lon, lat, pollon, pollat):
    """\
    Convert points using cartopy (much faster)
    """
    c_in  = ccrs.PlateCarree()
    c_out = ccrs.RotatedPole(pollon, pollat)

    if np.ndim(lon) == 0:
        res = c_out.transform_point(lon, lat, c_in)
        return res[0], res[1]
    else:
        res = c_out.transform_points(c_in, lon, lat)
        return res[...,0], res[...,1]


def spatial_binning(a, n=2):
    if np.ndim(a) == 2:
        kernel = np.ones((n,n)) / n**2
        return ndi.convolve(a, kernel)[::n,::n]
    else:
        kernel = np.ones((1,n,n)) / n**2
        return ndi.convolve(a, kernel)[:,::n,::n]


