
from six import string_types
from datetime import datetime, timedelta
import itertools
import os

from PIL import Image
from skimage import draw
from skimage import measure
from skimage import morphology as mo


import matplotlib.pyplot as plt
import numpy as np
import scipy.ndimage as ndi
import scipy.stats as stats
import scipy.optimize as opt
import pandas

from smartcarb import io
from smartcarb import map
from smartcarb import misc
from smartcarb import config as cfg

# location of Berlin
BERLIN = 13.407696, 52.516984


# ignore invalid floating-point operation
np.seterr(invalid='ignore')


BG_STATS_FILENAME = {
        'Berlin': {
            'CO2': '/project/smartcarb/ts/data/ts_XCO2_background_Berlin_100.dat',
            'NO2': '/project/smartcarb/ts/data/ts_NO2_background_Berlin_100.dat'
        },
        'Janschwalde': {
            'CO2': '/project/smartcarb/ts/data/ts_XCO2_background_Janschwalde_50.dat',
            'NO2': '/project/smartcarb/ts/data/ts_NO2_background_Janschwalde_50.dat'
        }
}


def get_ground_truths(filename, origin, tracer, thr, cloud_thr, dn):

    # lon/lat of source origin
    lon_o, lat_o = get_source_origin(origin)

    # time and ground pixels
    time = io.filename2time(filename)
    lon0, lat0, lonc, latc = io.read_ground_pixels(filename)

    # read signal and clouds
    signal_name = get_signal_name(tracer, origin)
    signal, clouds = io.read_fields(filename, [signal_name, 'CLCT'], time,
            correct_berlin=True)

    if thr is None:
        thr = {'CO2': 0.1, 'NO2': 0.2e15}[tracer]

    truths = find_plume_by_signal(signal, thr, lon_o, lat_o, lon0, lat0, dn, clouds, cloud_thr)


    # signal-weighted center of true plume
    lon_0 = np.sum(signal[truths]*lon0[truths]) / np.sum(signal[truths])
    lat_0 = np.sum(signal[truths]*lat0[truths]) / np.sum(signal[truths])


    return truths, lon_0, lat_0



def find_plume_by_signal(signal, thr, lon_o=BERLIN[0], lat_o=BERLIN[1],
        lon0=None, lat0=None, dn=10, clouds=None, cloud_thr=0.01,
        **kwargs):
    """\
    Find plume that has signal above threshold and is connected to location
    given by lon_o and lat_o.
    """
    if clouds is None:
        is_plume = signal > thr
    else:
        is_plume = (signal > thr) & (clouds < cloud_thr)

    labels = measure.label(is_plume, background=0)
    return find_plume_at_location(labels, lon_o, lat_o, lon0, lat0, dn)



def find_plume_at_location(labels, lon_o, lat_o, lon, lat, dn=10):

    try:
        s = get_area_around_source(lon_o, lat_o, lon, lat, dn)
    except ValueError:
        return np.zeros(labels.shape, bool)

    numbers = set(labels[s].flatten()) - {0}

    if numbers:
        return  np.any([labels == i for i in numbers], axis=0)
    else:
        return np.zeros(labels.shape, bool)


class ConfusionMatrix:
    def __init__(self, detection, truths):
        self.TP = np.sum(detection & truths)
        self.TN = np.sum(~detection & ~truths)
        self.FP = np.sum(detection & ~truths)
        self.FN = np.sum(~detection & truths)

        # the number of real positive cases in the data 
        self.P = self.TP + self.FN

        # true positive rate (sensitivity)
        self.TPR = self.TP / (self.TP + self.FN)

        # positive predictive value (precision)
        self.PPV = self.TP / (self.TP + self.FP)

        # F1-scoare (harmonic mean of TPR and PPV)
        self.F1 = 2 * self.TP / (2 * self.TP + self.FP + self.FN)


    def to_line(self, sep='\t', give_heading=False):
        if give_heading:
            return 'P\tTP\tTN\tFP\tFN\tTPR\tPPV\tF1\n'
        else:
            numbers = '\t'.join('%d' % v for v in [self.P, self.TP, self.TN, self.FP, self.FN])
            rates = '\t'.join('%.3f' % v for v in [self.TPR, self.PPV, self.F1])

            return numbers + '\t' + rates + '\n'





def label_plumes(d, n_min=20):
    """
    Label detected plume pixels.
    """
    d[np.isnan(d)] = 0
    labels = measure.label(d, background=0)

    i = 1
    final = np.zeros_like(labels)

    for l in set(labels.flatten()):
        if l != 0 and np.sum(labels == l) >= n_min:
            final[labels == l] = i
            i += 1

    return final


def welch_satterthwaite_equation(s1, s2, n1, n2):

    a = (s1**2 / n1 + s2**2 / n2)**2
    b = s1**4 / (n1**2 * (n1-1)) + s2**2 / (n2**2 * (n2-1))

    return a/b


def local_stats(img, size=9):

    if size == 5:
        footprint = np.array([
            [0, 1, 0],
            [1, 1, 1],
            [0, 1, 0]
        ])
    elif size in [9,16,25]:
        size = int(np.sqrt(size))
        footprint = np.ones((size,size))
    else:
        raise ValueError

    def nanmean(x):
        n = np.sum(np.isfinite(x))
        return np.nanmean(x) if n>0 else np.nan

    def nanvar(x):
        n = np.sum(np.isfinite(x))
        return np.nanvar(x, ddof=1) if n>1 else np.nan


    domain = np.any(img, axis=1)

    mean = np.full(img.shape, np.nan)
    mean[domain,:] = ndi.generic_filter(
            img[domain,:], nanmean, footprint=footprint, mode='constant', cval=np.nan
    )
    var = np.full(img.shape, np.nan)
    var[domain,:] = ndi.generic_filter(
            img[domain,:], nanvar, footprint=footprint, mode='constant', cval=np.nan
    )

    n = np.full(img.shape, np.nan)
    n[domain,:] = ndi.generic_filter(
            np.isfinite(img).astype(int)[domain,:], np.count_nonzero,
            footprint=footprint, mode='constant', cval=0
    )

    return mean, var, n





def get_area_around_source(lon_o, lat_o, lon, lat, bg_area):
    # define area of background around Berlin

    i,j = misc.nearest_index(lon_o, lat_o, lon, lat)

    d = (np.sqrt((lon[i,j] - lon_o)**2 + (lat[i,j] - lat_o)**2))

    if d > 0.1: # about 10 km
        raise ValueError('source (likely) outside swath')

    imin, imax = max(0, i-bg_area), i+bg_area
    jmin, jmax = max(0, j-bg_area), j+bg_area
    bg = slice(imin, imax), slice(jmin, jmax)

    return bg


def detect(lon, lat, field, is_cloudy, footprint,
        lon_o, lat_o, source_area=1,
        p_thr=0.997, sigma=None,
        bg_type='ts', bg_values=None, bg_area=50,
    ):
    """\
    Detect plume in `field` at location (`lon_o`, `lat_o`)

    Parameters
    ''''''''''
    lon :: np.ndarray
        longitudes at ground pixel centers

    lat :: np.ndarray
        latitudes at ground pixel centers

    field :: np.ndarray
        synthetic satellite observations (with noise)

    is_cloudy :: np.ndarray (boolean)
        array with True where clouds

    footprint :: int
        size of footprint used for computing local stats

    lon_o, lat_o
        longitude and latitude of source origin

    source_area
        size of source area (in pixels)

    p_thr :: float
        detection threshold 

    sigma :: float
        single sounding precision (only used if footprint == 1 and thus
        sample variance cannot be computed from neighborhood)

    bg_type

    bg_values :: np.array or None
        field used for computing mean and variance of background
        if None use field

    bg_area:: int
        size of background area around origin (lon_o, lat_o)

    origin

    tracer

    """

    # if footprint is even, center pixel is not defined clearly.
    if footprint % 2 == 0:
        print('Warning: footprint should be odd number (1,3,5,...).')

    if bg_type == 'ts':

        mean_b = bg_values['mean']
        var_b = bg_values['var']
        n_b = bg_values['n']

    elif bg_type in ['model', 'sat']:
        # estimate background mean and variance
        bg_slice = get_area_around_source(lon_o, lat_o, lon, lat, bg_area=bg_area)

        if bg_values is None:
            bg_values = field.copy()

        bg_values = bg_values[bg_slice]

        mean_b = np.nanmean(bg_values)
        var_b = np.nanvar(bg_values)

        n_b = np.isfinite(bg_values).sum()

    # estimate local mean and variance
    if footprint == 1:
        mean_s = field.copy()
        var_s = np.full(field.shape, sigma**2)
        n_s = np.ones_like(field)
    else:
        mean_s, var_s, n_s = local_stats(field, size=footprint)


    # Welch's t-test
    t_values = (mean_s - mean_b) / np.sqrt(var_s / n_s + var_b / n_b)

    # degree of freedom
    if footprint == 1:
        dof = np.ones_like(field)
    else:
        dof = welch_satterthwaite_equation(var_s, var_b, n_s, n_b)


    # t-value depends on degree of freedom (Welch's t-test)
    is_hit = find_hits(t_values, p_thr, dof)

    # label plumes
    labels = label_plumes(is_hit, 10) # TODO: de-hardcode

    is_plume = find_plume_at_location(labels, lon_o, lat_o, lon, lat, dn=source_area)

    # some pixels in `in_plume` might be detected (as neighborhood) but cloudy
    # these pixels will be set to false, because they are not part of ground
    # truth and cannot be used for flux inversion (however, they are used to
    # label regions)
    is_plume[is_plume & is_cloudy] = False


    info = {'mean': mean_s, 'var': var_s, 'number': n_s,
            'mean_bg': mean_b, 'var_bg': var_b, 'n_bg': n_b,
            'dof': dof,
            't_values': t_values, 'is_hit': is_hit,
            'labels': labels,
    }

    return is_plume, info


def get_source_origin(name):
    if name == 'Berlin':
        lon_o, lat_o = cfg.CITIES[name]
    else:
        lon_o, lat_o = cfg.POINT_SOURCES[name]

    return lon_o, lat_o


def get_signal_name(tracer, origin):

    suffix = 'BV' if origin == 'Berlin' else 'JV'

    if tracer == 'CO2':
        tracer = 'XCO2'

    return '%s_%s' % (tracer, suffix)




def plume_detection_algorithm(filename, tracer='CO2', error_std=0.7,
        cloud_thr=0.01, seed='orbit', use_epf=False, p_thr=0.99, footprint=3,
        bg_area=None, bg_type='ts',
        origin='Berlin', show=False):
    """\
    Origin is Berlin or others (i.e. "Janschwalde" tracer)
    """

    # lon/lat of source origin
    lon_o, lat_o = get_source_origin(origin)

    # time and groud pixels
    time = io.filename2time(filename)
    lon0, lat0, lonc, latc = io.read_ground_pixels(filename)

    # read signal, clouds and total tracer field
    signal_name = get_signal_name(tracer, origin)
    signal, clouds = io.read_fields(filename, [signal_name, 'CLCT'], time,
            correct_berlin=True)

    true_field = io.read_total_tracer(filename, tracer, time, correct_berlin=True)

    # read noise
    if seed == 'orbit':
        _, orbit, _, lon_eq = io.process_filename(filename)
        np.random.seed(orbit + (lon_eq // 90 * 6000))
    elif seed is not None:
        np.random.seed(seed)

    noise = io.read_tracer_noise(filename, tracer, use_epf, error_std)

    # field is an array without mask (masked values should be set to nan)
    field = np.array(true_field + noise)
    field[true_field.mask | (clouds > cloud_thr)] = np.nan

    # background values
    if bg_type == 'sat':
        bg_values = field.copy()
    elif bg_type == 'model':
        bg_values = field - signal
    elif bg_type == 'ts':
        fn = cfg.BG_STATS_FILENAME[origin][tracer]
        d = pandas.read_csv(fn, sep='\t', parse_dates=True, index_col=0)

        bg_values = {}
        bg_values['mean'] = float(d[d.index == time.strftime('%Y-%m-%d %H:%M')]['mean_bg'])
        bg_values['var'] = float(d[d.index == time.strftime('%Y-%m-%d %H:%M')]['std_bg']**2)
        bg_values['n'] = 72

    else:
        raise ValueError("bg_type has to be 'sat', 'model' or 'ts' not '%s'" % bg_type)


    # source area
    if origin == 'Berlin':
        source_area = 10 # (20x20 km2)
    elif origin == 'Lippendorf':
        source_area = 5
    else:
        source_area = 2  # 4x4 km2)

    if tracer == 'NO2':
        if error_std == 'low': # low noise scenario
            error_std = 1e15
        elif error_std == 'high': # high noise scenario
            error_std = 2e15

    # detect plume
    is_plume, info = detect(
            lon0, lat0, field, (clouds > cloud_thr), footprint,
            lon_o, lat_o, source_area,
            p_thr, error_std,
            bg_type, bg_values, bg_area,
    )


    info.update({
        'lon0': lon0, 'lat0': lat0,
        'lonc': lonc, 'latc': latc,
        'signal': signal, 'field': true_field,
        'noise':  noise,  'clouds': clouds,
        'lon_o': lon_o, 'lat_o': lat_o
    })

    # get ground truth (CO2 plume above 0.1 ppm with cloud fractions below 30%)
    info['truths'], lon_0, lat_0 = get_ground_truths(filename, origin, 'CO2', 0.1, 0.3, source_area)

    # confusion matrix for detecting the NO2 or CO2 plume
    info['C'] = ConfusionMatrix(is_plume, info['truths'])

    if show:
        visualize(tracer, is_plume=is_plume, do_zoom=True, show_clouds=True,
                use_noise=True, marker='+', cloud_thr=cloud_thr, p_thr=p_thr,
                footprint=footprint, lon_0=lon_0, lat_0=lat_0, **info)

    return is_plume, info





def find_hits(t_values, p_thr, dof):
    t_thr = stats.t.ppf(p_thr, dof)
    is_hit = t_values > t_thr
    return is_hit


def find_contour(lon, lat, binary):
    for contour in  measure.find_contours(binary, 0.0):
        coords = measure.approximate_polygon(contour, tolerance=1.0)

        n,_ = np.shape(coords)
        if n <= 4:
            continue

        i,j = coords[:,0].astype(int), coords[:,1].astype(int)
        yield lon[i,j], lat[i,j]





def visualize(tracer, lonc, latc, signal, field, noise, clouds, is_plume, labels,
        is_berlin=None, show_clouds=True, do_zoom=True, lon0=None, lat0=None,
        lon_o=None, lat_o=None, lon_0=None, lat_0=None,
        use_noise=True, cloud_thr=None, truths=None,
        p_thr=None, footprint=None, mean_bg=None, **kwargs):

    # tracer field
    if not use_noise:
        noise = 0.0

    if not show_clouds:
        clouds = None

    if clouds is not None and cloud_thr is not None:
        clouds = clouds > cloud_thr

    if tracer == 'CO2':
        if mean_bg is None:
            mean_bg = field.mean()
        fig = map.make_level2_map(
            lonc, latc, field - mean_bg + noise, vmin=-2, vmax=2,
            label='XCO$_2$ signal (ppm)', clct=clouds,
        )
    else:
        fig = map.make_level2_map(
            lonc, latc, (field + noise) / 1e15, vmin=0, vmax=20, truncate_cmap=True,
            label='NO$_2$ column (10$^{15}$ molecules per cm$^2$)', clct=clouds,
        )
    #
    ax, cax = fig.get_axes()
    map.add_city_labels(ax)


    true_line = []
    if truths is not None and np.sum(truths) > 0:
        #truths = mo.remove_small_holes(truths)
        for i,(x,y) in enumerate(find_contour(lon0, lat0, truths)):
            true_line = ax.plot(x,y, 'k--', transform=map.ccrs.PlateCarree())


    if is_plume is not None and np.sum(is_plume) > 0:
        detected_line = ax.plot(lon0[is_plume], lat0[is_plume], marker='o', mec='none', ms=1.5, color='k',
                ls='', transform=map.ccrs.PlateCarree())
        #for i,(x,y) in enumerate(find_contour(lon0, lat0, is_plume)):
        #   detected_line = ax.plot(x,y, 'k', dashes=(2,1), transform=map.ccrs.PlateCarree())
    else:
        detected_line = []

    if do_zoom:
        if lon_0 is None or lat_0 is None:
            rlon_o, rlat_o = misc.wgs84_to_rotpole(lon_o, lat_o, cfg.pollon, cfg.pollat)
        else:
            rlon_0, rlat_0 = misc.wgs84_to_rotpole(lon_0, lat_0, cfg.pollon, cfg.pollat)

        dy = 1.1
        dx = dy * 1.12
        xmin, xmax = rlon_0 - dx, rlon_0 + dx
        ymin, ymax = rlat_0 - dy, rlat_0 + dy

        ax.set_xlim(xmin, xmax)
        ax.set_ylim(ymin, ymax)


    lines = ax.plot(cfg.pollon, cfg.pollat, marker='s', mec='k', mfc='w', ls='',
            transform=map.ccrs.PlateCarree())
    labels = ['cloud fraction > %d%%' % (100*cloud_thr)]

    if true_line:
        lines += true_line
        labels += ['"true" plume (XCO$_2$ > 0.1 ppm and cloud fraction < 30%)']

    if detected_line:
        lines += detected_line
        labels += ['detected plume (p$_{thr}$ = %.2f, neighborhood = %d)' % (p_thr, footprint)]

    if lines:
        ax.legend(lines, labels, loc="upper left", numpoints=1, markerscale=1.5)


def show_all_ground_truths(filename, thr=0.1, cloud_thr=0.01, dn=10,
        output_path=None):

    # time and groud pixels
    time = io.filename2time(filename)
    lon, lat, lonc, latc = io.read_ground_pixels(filename)

    bv,jv,clct = io.read_fields(filename, ['XCO2_BV', 'XCO2_JV', 'CLCT'])

    names = ['Janschwalde', 'Schwarze Pumpe', 'Boxberg',
            'Lippendorf', 'Turow', 'Patnow']
    colors = [
        'grey',
        '#1f77b4', # Berlin
        '#ff7f0e', # J
        '#2ca02c', # SP
        '#d62728', # Bb
        '#9467bd', # Ld
        '#8c564b', # T
        '#e377c2', # P
        #'#7f7f7f',
        '#bcbd22',
        #'#17becf',
    ]
    labels = [
        'back-\nground',
        'Berlin',
        'Jänsch-\nwalde',
        'Schwarze\nPumpe',
        'Boxberg',
        'Lippen-\ndorf',
        'Turów',
        'Patnów',
        'ambig-\nuous'
    ]

    # init mask
    mask = np.zeros(lon.shape, dtype=int)

    # Berlin plume
    lon_o, lat_o = cfg.CITIES['Berlin']
    p = find_plume_by_signal(bv, thr, lon_o, lat_o, lon, lat, dn=dn,
                            clouds=clct, cloud_thr=cloud_thr)
    mask[p] = 1

    # power plant plumes
    shape = mask.shape + (6,)
    plants = np.zeros(shape, dtype=int)

    for i, name in enumerate(names, 2):
        lon_o, lat_o = cfg.POINT_SOURCES[name]
        p = find_plume_by_signal(jv, thr, lon_o, lat_o, lon, lat, dn=dn,
                            clouds=clct, cloud_thr=cloud_thr)

        plants[p,i-2] = i
        mask[p] = i

    # ambiguous plunmes
    mask[np.sum(plants > 0, axis=2) > 1] = len(names)+2

    labels2 = [labels[0]]
    for i, label in enumerate(labels[1:], 1):
        label = '%s\n[%d]' % (label, np.sum(mask == i))
        labels2.append(label)
    labels = labels2

    # nothing detected
    if not np.any((mask != 0) & (clct < cloud_thr)):
        return None

    # reverse numbering, colors and labels
    mask = -1 * mask + len(names) + 2
    colors = colors[::-1]
    labels = labels[::-1]

    # plot
    fig = map.make_level2_map(
        lonc, latc, mask, vmin=None, vmax=None,
        label='', clct=clct, is_discrete=True, cmap=colors,
        cb_labels=labels
    )

    ax, cax = fig.get_axes()
    map.add_city_labels(ax, size='medium')

    text = io.filename2label(filename)
    ax.text(0.0, 1.0, text, ha='left', va='top',
             bbox=dict(facecolor='white', alpha=1.0),
             transform=ax.transAxes
    )

    text = 'XCO$_2^\mathrm{thr}$ = %.2f ppm, d$_\mathrm{max}$ = %d px' % (thr, dn)
    ax.text(0.01, 0.0, text, ha='left', va='bottom', fontsize='x-small',
             transform=ax.transAxes
    )


    if output_path is not None:
        filename = os.path.basename(filename)
        filename = filename.replace('.nc', '_plumes.png')
        filename = os.path.join(output_path, filename)

        plt.savefig(filename)

    return True



def show_ground_truth(lonc, latc, signal, clouds, do_zoom=True, show_clouds=True,
        thr=0.1, cloud_thr=0.01, mask_clouds=True, lon_o=BERLIN[0], lat_o=BERLIN[1],
        dn=10, **kwargs):

    if mask_clouds:
        mask = find_plume_by_signal(signal, thr, lon_o, lat_o,
                lonc.mean(2), latc.mean(2), dn=dn,
                clouds=clouds, cloud_thr=cloud_thr)
    else:
        mask = find_plume_by_signal(signal, thr, lon_o, lat_o,
                lonc.mean(2), latc.mean(2), dn=dn)

    n_pixels = np.sum(mask)

    mask = (~mask).astype(int)
    if show_clouds:
        mask[clouds > cloud_thr] = 2

    fig = map.make_level2_map(
        lonc, latc, mask, vmin=None, vmax=None,
        label='', clct=None, is_discrete=True, cmap='Greys_r',
        cb_labels=['plume\n[%d]' % n_pixels, 'back-\nground', 'clouds']
    )
    ax, cax = fig.get_axes()
    map.add_city_labels(ax)

    if do_zoom:
        ax.set_xlim(1, 3.8)
        ax.set_ylim(3.5, 6)

    text = 'XCO$_2^\mathrm{thr}$ = %.2f ppm, d$_\mathrm{max}$ = %d px' % (thr, dn)
    ax.text(0.01, 0.0, text, ha='left', va='bottom', fontsize='x-small',
             transform=ax.transAxes
    )


def show_labels(lonc, latc, labels, clouds, do_zoom=True, **kwargs):
    # labels
    fig = map.make_level2_map(
        lonc, latc, labels, vmin=None, vmax=None,
        label='labels of detected plumes', clct=clouds > 0.01, is_discrete=True,
        cmap='nipy_spectral_r'
    )
    if do_zoom:
        ax, cax = fig.get_axes()
        ax.set_xlim(1, 3.8)
        ax.set_ylim(3.5, 6)
        map.add_city_labels(ax)


def show_plume(lonc, latc, is_plume, clouds, do_zoom=True, **kwargs):

    mask = np.ones(is_plume.shape, int)
    mask[is_plume] = 0
    mask[clouds > 0.01] = 2

    fig = map.make_level2_map(
        lonc, latc, mask, vmin=None, vmax=None,
        label='', clct=None, is_discrete=True, cmap='Greys_r',
        cb_labels=['plume', 'back-\nground', 'clouds'],
    )
    if do_zoom:
        ax, cax = fig.get_axes()
        ax.set_xlim(1, 3.8)
        ax.set_ylim(3.5, 6)
        map.add_city_labels(ax)





def plot_performance():

    cases = [
        ('NO2', 'high', 0.30),
        ('NO2', 'high', 0.01),
        ('NO2', 'low', 0.30),
        ('NO2', 'low', 0.01),
        ('CO2', 1.0, None),
        ('CO2', 0.7, None),
        ('CO2', 0.5, None),
    ]
    data = {}
    products = []


    labels = [
        #r'NO$_2$ ($\sigma_{ref}$ = 2$\times$10$^{15}$ cm$^{-2}$, c < 30$\%$)', # high
        #r'NO$_2$ ($\sigma_{ref}$ = 2$\times$10$^{15}$ cm$^{-2}$, c < 1$\%$)',  # high
        #r'NO$_2$ ($\sigma_{ref}$ = 1$\times$10$^{15}$ cm$^{-2}$, c < 30$\%$)', # low
        #r'NO$_2$ ($\sigma_{ref}$ = 1$\times$10$^{15}$ cm$^{-2}$, c < 1$\%$)',  # low
        #r'CO$_2$ ($\sigma_{ref}$ = 1.0 ppm)',
        #r'CO$_2$ ($\sigma_{ref}$ = 0.7 ppm)',
        #r'CO$_2$ ($\sigma_{ref}$ = 0.5 ppm)',
        r'NO$_2$ HW',     # high
        r'NO$_2$ HS',  # high
        r'NO$_2$ LW', # low
        r'NO$_2$ LS',  # low
        r'CO$_2$ H',
        r'CO$_2$ M',
        r'CO$_2$ L',
    ]

    rate2long_name = {
        'TPR': 'true-positive rate',
        'PPV': 'positive predictive value',
        'F1': 'F1 score'
    }


    for tracer, error_std, cloud_thr in cases:

        if tracer == 'CO2':
            product = 'CO2_%02d' % (10 * error_std)
        else:
            product = '%s_%s_cc%d' % (tracer, error_std, 100*cloud_thr)

        path = '/project/smartcarb/plume_detection/'
        filename = os.path.join(path, 'confusion_matrix_%s_Berlin_p0.95_n5.dat' % product)

        products.append(product)
        data[product] = pandas.read_csv(filename, sep='\t')

    fig = plt.figure(figsize=(10,3))

    for i,rate in enumerate(['TPR', 'PPV', 'F1']):
        ax = fig.add_subplot(1,3,i+1)

        # remove nans (e.g. if TP == 0)
        values = [data[p][rate] for p in products]
        values = [v[np.isfinite(v)] for v in values]

        ax.boxplot(values, whis=np.inf, sym='', vert=False)
        ax.set_title('(%s) %s' % ('abc'[i], rate2long_name[rate]))
        ax.set_xlim(-0.01,1.01)

        if i == 0:
            ax.set_yticklabels(labels, rotation=0, fontsize='small')
        else:
            ax.set_yticklabels([])

    plt.tight_layout()



def main(output_path, level2_path, tracer, error_std,
        origin='Berlin', p_thr=0.95, footprint=5, cloud_thr=0.01):

    level2_filenames = np.loadtxt(
            '/home/kug/projects/smartcarb/plume_detection/Level-2_%s.dat' % origin,
            dtype=bytes
        ).astype(str)


    if tracer == 'CO2':
        product = 'CO2_%02d' % (10 * error_std)
    else:
        product = '%s_%s_cc%d' % (tracer, error_std, 100*cloud_thr)

    output_filename = 'confusion_matrix_%s_%s_p%.2f_n%d.dat' % (product, origin, p_thr, footprint)
    output_filename = os.path.join(output_path, output_filename)

    with open(output_filename, 'w') as out:

        for i, filename in enumerate(level2_filenames):
            print(filename, product)

            is_plume, info = plume_detection_algorithm(os.path.join(level2_path, filename),
                    tracer=tracer, use_epf=True, error_std=error_std,
                    bg_type='ts',
                    origin=origin,
                    p_thr=p_thr, footprint=footprint, cloud_thr=cloud_thr,
                    show=True
            )

            # savefig and close
            plt.savefig(
                os.path.join(output_path, filename.replace('.nc', '%s.png' % product))
            )
            plt.savefig(
                os.path.join(output_path, filename.replace('.nc', '%s.pdf' % product))
            )
            plt.close()


            # write confusion matrix to files
            C = info['C']

            if i == 0:
                out.write(C.to_line(give_heading=True))

            out.write(C.to_line())
            out.flush()


if __name__ == '__name__':
    pass







